#include "type_base.h"
#include "gin_cst.h"
#include "config.h"
#include "spectre.h"
#include "vector.h"

typedef Flt Vec[3] ;
typedef Vec Point ;
typedef Flt Matrix[4][4] ;

#define max(a,b) 	((a)>(b)?(a):(b))
#define min(a,b) 	((a)<(b)?(a):(b))
#define SQR(a)          ((a)*(a))

/*----------------------------------------------------------------------*/

typedef Flt Rprecalcule[MAXANGLES];

typedef struct 
{
    Flt n, k, F0moyLu, F0moyC;
    int opaque;
    spectral F0;
    Rprecalcule Rtlm;
}
Tparamphys;

#define INDICSIZE 10
typedef Flt indic[INDICSIZE];

typedef struct t_lumin 
{
    indic indicl1, indicl2, indict1,indict2;
    Vec   U,V,N;
    Flt   locabs[4][4];
    Flt   absloc[4][4];
    Flt   S, lumens, nits;
    MazdaSpec mazdaspec;
}
Indicat;

typedef struct t_surface {
        int     paraphys;
	Flt	surf_s ;
	Flt	surf_d ;
	Flt	surf_m ;
	Flt     surf_lum;
	boolean source;
	Indicat *indics;
} Surface ;

/*----------------------------------------------------------------------*/

typedef char Tnommat[LGID];

typedef struct
{
    double wl,val;
}
elntcourbe;

typedef elntcourbe Tcourbe[WLMAX-WLMIN+1];

typedef struct t_light {
        Vec	        light_pos ;
	spectral	light_brightness ;
	struct t_object * light_obj_cache[MAXLEVEL] ;
} Light ;

typedef struct t_viewpoint {
	Vec	view_from ;
	Vec	view_at ;
	Vec	view_up ;
	Flt	view_angle ;
	Flt	view_dist ;
} Viewpoint ;

typedef struct t_object {
	unsigned short 	o_type ;
	Flt	o_dmin[NSLABS] ;
	Flt	o_dmax[NSLABS] ;
	struct t_objectprocs {
		int 	(*print) () ;
		int 	(*intersect) () ;
		int 	(*normal) () ;
	} * o_procs ;
	struct   t_surface 	* o_surf ;
	void	 * o_data ;
	int      nsamples;
	int      * sample;
	int      m,n;
	int      udepth,vdepth;
	spectral *texture;
	spectral *map;
} Object ;

typedef struct t_compositedata {
	unsigned short 	c_size ;
	Object *	c_object[BUNCHINGFACTOR] ;
} CompositeData ;

typedef struct t_objectprocs ObjectProcs ;

typedef struct t_isect {
	Flt 		isect_t ;
	int 		isect_enter ;
	Object 		* isect_prim ;
	Surface 	* isect_surf ;
	Vec             isect_normal ;
} Isect ;

typedef struct t_inout {
	unsigned char gauche;
        unsigned char in;
        Isect *inter;       
        struct t_inout *next;  
} inout ; 

typedef struct t_pixel {
	unsigned char r, g, b, q ;
} Pixel ;

typedef struct t_noeud { 
                char type_op;
                Flt o_dmin[NSLABS];
                Flt o_dmax[NSLABS];
} noeud;

#define Opens  2
#define Solid 1
 

#ifndef PI
#define PI 		(3.14159265358979323844)
#endif /* PI */

#define degtorad(x)	(((Flt)(x))*PI/180.0)

#define		T_COMPOSITE	(0)
#define		T_SPHERE	(1)
#define		T_POLY		(2)
#define		T_CONE		(3)
#define         T_TRI           (4)
#define         T_CYLINDER      (5)
#define         T_BOX           (6)

#ifdef __GNUC__
#define INLINE	inline

#ifdef FAST_MATH_PRIMS

#define sqrt(x)\
	({ double __value, __arg = (x) ;\
	 asm("fsqrtx %1,%0": "=f" (__value): "f" (__arg)) ;\
	 __value ;})

#endif /* FAST_MATH_PRIMS */

#else /* __GNUC__ */

#define INLINE	/* */

#endif /* __GNUC__ */



/******************************************************************************

 Declaration des types utilises par le logiciel GIN

******************************************************************************/

typedef struct
{
    
    Flt u,v;
}
CoordLoc;

typedef struct
{
    Point    CoordG;
    CoordLoc CoordL;
    CoordLoc CoordLmin;
    CoordLoc CoordLmax;
    int      udepth,vdepth;
    Vec      normale;
    Flt      area;
    Object   *obj;
    spectral lgd;
    spectral dlgd;
    entier   indsi[3];
}
SplePt; /* sample point */

typedef enum {aucune,partielle,totale} inter_box;

typedef struct tboite
{
    /* position et taille ... */

    int      nsamples;
    int      * sample;
    int      nobjs;
    int      * obj;
    Point    center;
    Flt      bound;
    int      niveau[3], adresse[3];
    Vec      pos_min, pos_max;
}
boite;

typedef boite * Aboite;

typedef unsigned char *graph;      

typedef struct
{
    boolean visible;
    graph   subgraph;
    entier  size;
    Object  *obj;
}
visibilite;


typedef struct tcone
{
    Vec   dir;
    Point sommet;
    Flt   ouverture,tanouv;
}
cone;

typedef struct t_polydata {
  int 	poly_npoints ;
  Vec	* poly_point ;
  Vec	poly_normal ;
  Flt 	poly_d ;
  Flt	poly_p1, poly_p2 ;
  int     n,m;
} PolyData ;

typedef struct t_spheredata {
  Vec 		sph_center ;
  Flt 		sph_radius ;
  Flt 		sph_radius2 ;
  Flt           sph_area;
} SphereData ;

/*----------------------------------------------------------------------*/

typedef struct
{
	entier   p1,p2;        /* numeros des points depart, arrivee */
	Point    P1,P2;        /* coordonnees (x,y,z) des points */
	boite    *B1,*B2;      /* pointeurs boites arrivee, depart */
	entier   b1, b2;       /* numeros boites arrivee, depart */
	entier   pos1,pos2;    /* adresse des pts dans leur boite (obj?) */
	Object   *obj1, *obj2; /* objets depatr, arrivee */
	Point    P;            /* pour compatibilite avec ancienne version */
	Vec      D;            /* vecteur directeur P1P2 */
	Flt      length;       /* distance P1P2 */
	Flt      t;            /* distance parcourue depuis P1 */
	Flt      ang1,ang2;    /* cos angle rayon et normales en P1 et P2 */
	Object   *lightbuf;    /* objet susceptible d'intercepter le rayon */
	Flt      solang;       /* angle solide = (S1*ang1*ang2)/SQR(length) */
	entier   indsi[3];     /* adresse spartiale de P1 */
}
Ray ;

typedef Ray SolidAngle;
typedef Ray ShadowRay;

#define RayPoint(ray,t,point)	VecAddS(t,(ray)->D,(ray)->P,point)

typedef struct
{
    entier ind,pos;
    boite  *box;
}
SpleAdr;
