
/****************************************************************/
/*                                                              */
/* nff2dat.c :                                                  */
/*             converts the nff format into a dat format readable
               by the radiosity program of Pattanaik            */
/*                                                              */
/*  nff2dat : cc -o nff2dat nff2dat.c vec_util.c -lm            */
/****************************************************************/

#include <stdio.h>
#include <math.h>
#include "vec_util.h"

#define rindex         strrchr

#define SPECSIZE   4
#define WLMAX           780
#define WLMIN           380
#define BUFSIZE  128
#define DEFAULT_PATCH_MAX 50.0
#define PI 3.14159265358979323844

/* longueurs d'onde selectionnees */
double wl[SPECSIZE+1]={456.4, 490.9, 557.7, 631.4, 589.0};

#define quadrilateral 0
#define Sphere        1
#define Bubble        2
#define Cylinder      3
#define Tube          4
#define Cone          5
#define Cup           6
#define Ring          7
#define Polygon       8
#define others        9

#define true  1
#define false 0

typedef double spectral[SPECSIZE];
typedef struct { double wl,val;} elntcourbe;
typedef elntcourbe Tcourbe[WLMAX-WLMIN+1];
typedef double Flt;
typedef Flt Vec[3];


FILE *fin, *fout, *finter;

char nomf[16],nfin[16],nfout[16],nfinter[16];
char nomspectre[20],distrib[20];
char buffer[BUFSIZE];
char commande[80];

Vec  view_from,view_at,view_up;
Vec  v, n, p1, p2,vue, vue_from, vue_at, vue_up;
Vec  vert4[4];

spectral spectre;

double d, m, puissance;
double r1, r2;
double angle,hither; 
double bg;
double taille_patch_max;

int Xresolution,Yresolution;
int num_vert, tot_vert;
int emitter,specular;
int typobj;
int hd=0,vd=0;
int nPrims=0,nSamples=0;
int line = 0;

/*------------------------------------ functions -------------*/
 
int compwl(a, b)
     elntcourbe *a, *b;
{
    return( (int)(a->wl - b->wl));
    /* OK si echantillons distants d'au moins 1 nm */
}

void calc_hd_vd_quadrilateral(vert4)
   Flt vert4[4][3];
   
{
   Vec v0,v1,v2,v3;
   int i;
      
   for (i=0;i<3;i++) 
   {
      v0[i] = vert4[1][i] - vert4[0][i];
      v1[i] = vert4[2][i] - vert4[1][i];
      v2[i] = vert4[3][i] - vert4[2][i];
      v3[i] = vert4[0][i] - vert4[3][i];
    }
    hd =  (int)(NormeVecMin(v1,v3) / taille_patch_max);
    vd =  (int)(NormeVecMin(v0,v2) / taille_patch_max);   
    hd++;vd++;
    nSamples += hd*vd;
}

void calc_hd_vd_sphere(rayon)
{
    hd=vd=(int)((2*PI*rayon*rayon)/taille_patch_max);
    nSamples += hd*vd;
}
void calc_hd_vd_polygon()
{
    hd=vd=0;
}

void echantMeyer(tab, nb, wl, spect, f0)
     Tcourbe tab;
     int nb;
     double wl[SPECSIZE+1];
     spectral spect;
     double  *f0;
{
    int i,j;
    double         lambda;
    double spec[SPECSIZE+1];

    for (i=0;i<SPECSIZE+1;i++) {
        lambda = wl[i];
        if (lambda <= tab[0].wl)
        spec[i]=tab[0].val-(tab[1].val-tab[0].val)*(tab[0].wl-lambda)
        / (tab[1].wl - tab[0].wl);
        else {
            if (lambda >= tab[nb].wl)
            spec[i]=tab[nb-1].val+(tab[nb].val-tab[nb-1].val)
            *(lambda-tab[nb].wl)/(tab[nb].wl-tab[nb-1].wl);
            else {  /* recherche des valeurs qui encadrent lambda */
                for (j=0;j<=nb;j++)
                if (tab[j].wl > lambda)
                break;
                spec[i]=tab[j-1].val+(tab[j].val-tab[j-1].val)
                *(lambda-tab[j-1].wl)/(tab[j].wl - tab[j-1].wl);
            }
        }
    }
    
    (void)memcpy((char *)spect, (char *)spec, sizeof(spectral));
    *f0 = spec[SPECSIZE];
}
int CLR_read_mtl (file, opaque, n, k, courbe, nbval)
     char        *file;
     int         *opaque;
     double      *n, *k;
     Tcourbe     courbe;
     int         *nbval;
{
    FILE        *fp, *fopen();
    char        in_str[200];
    char        dir[80];
    char        name[80];
    int         nb = 0;


    sprintf(name,"/home/prompto/d4/siames/tellier/bdphoto/%s",file);
    

    if ((fp=fopen(name,"r")) == NULL) return false;
    if (n != NULL) *n = 0.0;
    if (k != NULL) *k = 0.0;
    if (opaque != NULL) *opaque = true;
    while (fgets(in_str,200,fp)) {
        switch (in_str[0]) {
          case '#':       /* comment */
            break;
          case 'k':       /* absorption coefficient */
            if (k != NULL) (void)sscanf (in_str, "k %lf", k);
            break;
          case 'n':       /* index of refraction */
            if (n != NULL) (void)sscanf (in_str, "n %lf", n);
            break;
          case 'O':       /* conductor */
            if (opaque != NULL) *opaque = true;
            break;
          case 'T':       /* dielectric */
            if (opaque != NULL) *opaque = false;
            break;
          default:        /* spectral data */
            if ((sscanf(in_str,
                        "%lf %lf", &courbe[nb].wl, &courbe[nb].val) == 2))
            nb++;
        }
    }
    if((k!=NULL)&&(*k!=0.0)) *opaque=true;
    *nbval = nb;
    (void)qsort((char *)courbe, nb, sizeof(elntcourbe), compwl);
    (void)fclose(fp);
    return true;
}

int read_material(name,curve)
     char *name;
     spectral curve;
{
    double      n,k,F0lu;
    Tcourbe  courbe;

    int      opaque;
    int      nbval;
 
    if (CLR_read_mtl (name, &opaque, &n, &k, courbe, &nbval))
    {
        echantMeyer(courbe, nbval-1, wl, curve, &F0lu);
    }
    return 1;
}

void write_param()
{
    int i;
    
    fprintf(finter,"%d %d\n",hd,vd);
    fprintf(finter,"        %d\n                ",emitter);
    if (emitter) 
       for(i=0; i<SPECSIZE;i++)
          fprintf(finter,"%lf ",spectre[i]*puissance);
    fprintf(finter,"\n");       
    fprintf(finter,"        %d\n                ",specular);
    for(i=0; i<SPECSIZE;i++)
       fprintf(finter,"%lf ",spectre[i]);
    fprintf(finter,"\n");
    fprintf(finter,"        %d\n",typobj);
}

int main(argc, argv)
     int argc ;
     char * argv[] ;
{
    extern char *optarg;
    extern char *rindex() ;
    extern int atoi();
    char *Progname ;
    char *help= "-i nff_file(.nff) [-t taille_patch_max][-o out_file]";
    int c;
    
    if ((Progname = rindex(argv[0], '/')) == NULL) 
    Progname = argv[0] ;
    else 
    Progname ++ ;
    if (argc == 1)
    {
	fprintf(stderr, "usage: %s %s\n", Progname,help);
	exit(0);
    }
    fprintf(stderr,"\n");

    /*-------------------- lecture des options -------------- */
    taille_patch_max=DEFAULT_PATCH_MAX;
    strcpy(nfout,"");
    while ((c = getopt(argc, argv, "i:o:t:")) != -1)
    {
	switch (c)
	{
	  case 'i':
	    strcpy(nomf, optarg);
	    break;
	  case 'o':
	    strcpy(nfout, optarg);
	    break;
	  case 't':
	    taille_patch_max = atof(optarg);
	    break;
	  case '?':
	    fprintf(stderr, "usage: %s %s\n", Progname,help);
	    exit(0);  
	 }
    } 
       
    strcpy(nfin,nomf);
    strcat(nfin,".nff");
    if (!strcmp(nfout,""))  strcpy(nfout,nomf);
    strcat(nfout,".dat");
    strcpy(nfinter,nomf);
    strcat(nfinter,"2.dat");
    

    if ((fin = fopen(nfin, "r")) == NULL)
	fprintf(stderr,"failed to open input file %s for reading\n",nfin);

    if ((fout = fopen(nfout, "w")) == NULL)
	printf(stderr,"failed to open output file for writing\n");
    
    if ((finter = fopen(nfinter, "w+")) == NULL)
       printf(stderr,"failed to open intermediate file for writing\n");
         
    while ( fgets( buffer, BUFSIZE, fin ) )
    {
	line++;
	/* POLYGONAL PATCH===================================== */
	if ( sscanf( buffer, "pp %d", &tot_vert ) == 1 )
	{
	    typobj = others;
	    fprintf(stderr,"polygonal patch  not implemented\n");
	    for (num_vert = 0; num_vert < tot_vert; ++num_vert )
	    {
		fgets( buffer, BUFSIZE, fin ) ;
		line++;
		sscanf(buffer,"%lf %lf %lf %lf %lf %lf",
		       &v[0],&v[1],&v[2],&n[0],&n[1],&n[2]);
	    }
	}
	/* POLYGON============================================= */
	else if ( sscanf( buffer, "p %d", &tot_vert ) == 1 )
	{

	    if (tot_vert == 4)
	    {
	        typobj = quadrilateral;
	        for (num_vert = 0; num_vert < tot_vert; ++num_vert )
	        {   
		   fgets( buffer, BUFSIZE, fin );
		   line++;
		   sscanf(buffer,"%lf %lf %lf",&vert4[num_vert][0],
		         &vert4[num_vert][1],&vert4[num_vert][2]);
	        }
	        calc_hd_vd_quadrilateral(vert4);
                write_param();
                for (num_vert = 0; num_vert < tot_vert; ++num_vert )
	        {
                fprintf(finter, "                %lf %lf %lf\n", 
                         vert4[num_vert][0],
                         vert4[num_vert][1], 
                         vert4[num_vert][2]);
            }

	    }
	    else
	    {
		typobj = Polygon;
	        calc_hd_vd_polygon(tot_vert);
	        write_param();
                fprintf(finter,"%d\n",tot_vert);
                for (num_vert = 0; num_vert < tot_vert; ++num_vert )
	        {
  		    fgets( buffer, BUFSIZE, fin );
		    line++;
		    sscanf(buffer,"%lf %lf %lf",&v[0],&v[1],&v[2]);
	      	    fprintf(finter, "%lf %lf %lf\n", v[0], v[1], v[2]);
                }
	    }
	    nPrims++;
	}
	/* SPHERE============================================== */
	else if ( sscanf( buffer, "s %lf %lf %lf %lf",
			 &p1[0], &p1[1], &p1[2], &r1 ) == 4 )
	{
	    typobj = Sphere;
	    calc_hd_vd_sphere(r1);
	    write_param();
	    fprintf(finter, "%lf %lf %lf %lf\n", p1[0], p1[1], p1[2], r1);
	    nPrims++;
	}
	/* CYLINDER / CONE===================================== */
	else if ( buffer[0] == 'c' )
	{
	    typobj = Cone;
	    fgets( buffer, BUFSIZE, fin ) ;
	    line++;
	    sscanf( buffer, "%lf %lf %lf %lf", 
		   &p1[0], &p1[1], &p1[2], &r1 ) ;
	    fgets( buffer, BUFSIZE, fin ) ;
	    line++;
	    sscanf( buffer, "%lf %lf %lf %lf", 
		   &p2[0], &p2[1], &p2[2], &r2 ) ;
	    write_param();
	    fprintf(finter,"%lf %lf %lf\n", p1[0], p1[1], p1[2]);
	    fprintf(finter,"%lf\n",r1);
	    fprintf(finter,"%lf %lf %lf\n", p2[0], p2[1], p2[2]);
	    fprintf(finter,"%lf\n",r2);
	    nPrims++;
	}
	/* PHOTOMETRIC PARAMETERS ============================== */
	else
	if (sscanf(buffer, "f %s %lf %lf %lf", nomspectre,&d,&m,&puissance) == 4)
	{
	    if (read_material(nomspectre, spectre) == 0)
	    {
		fprintf(stderr,"cannot find spectrum %s\n",nomspectre);
		exit(-1); 
	    }
	    if (d == 0.0) specular = true;
	    else specular = false;
	    if (puissance > 0.0) {
	       emitter = true;
	       specular = false;
	    }
	    else emitter = false;
	}
	/* ILLUMINATION PARAMETERS ============================== */
	else
	if (sscanf(buffer, "i %s %s %lf", nomspectre, distrib, &puissance) == 3)
	{
	    if (read_material (nomspectre, spectre) == 0)
	    {
		fprintf(stderr,"cannot find spectrum %s\n",nomspectre);
		exit(-1); 
	    }
	    if (d == 0.0) specular = true;
	    else specular = false;
	    if (puissance > 0.0) emitter = true;
	    else emitter = false;
	}
	/* POINT LIGHT    ============================================ */
	else if (sscanf(buffer,"l %lf %lf %lf",&p1[0],&p1[1],&p1[2]) == 3)
	{
	    fprintf(stderr,"light ignored\n");
	}
	/* VIEW PARAMETERS ===================================== */
	else if ( buffer[0] == 'v' )
	{
	    fgets( buffer, BUFSIZE, fin ) ;
	    line++;
	    sscanf(buffer,"from %lf %lf %lf",
		   &vue_from[0],&vue_from[1],&vue_from[2]);
	    fgets( buffer, BUFSIZE, fin ) ;
	    line++;
	    sscanf( buffer,"at %lf %lf %lf",&vue_at[0],&vue_at[1],&vue_at[2]);
	    fgets( buffer, BUFSIZE, fin ) ;
	    line++;
	    sscanf( buffer,"up %lf %lf %lf",&vue_up[0],&vue_up[1],&vue_up[2]);
	    fgets( buffer, BUFSIZE, fin ) ;
	    line++;
	    sscanf( buffer, "angle %lf", &angle ) ;
	    fgets( buffer, BUFSIZE, fin ) ;
	    line++;
	    sscanf( buffer, "hither %lf", &hither ) ;
	    fgets( buffer, BUFSIZE, fin ) ;
	    line++;
	    sscanf( buffer, "resolution %d %d\n", &Xresolution, &Yresolution);
	    fprintf(fout,"v\n");
	    fprintf(fout,"%d %d\n", Xresolution, Yresolution);
	    fprintf(fout,"%lf %lf %lf\n",vue_from[0],vue_from[1],vue_from[2]);
	    
	     /*((vue_at[0]==0)&&(vue_at[1]==1)&&(vue_at[2]==0));*/
	    VecSub(vue_at,vue_from,vue);
	    VecNormalize(vue);
	    fprintf(fout,"%lf %lf %lf\n",vue[0],vue[1],vue[2]);
	    fprintf(fout,"%lf %lf %lf\n",vue_up[0],vue_up[1],vue_up[2]);
	    fprintf(fout,"%lf %lf\n",angle,angle);
	    fprintf(fout,"%d\n",SPECSIZE);
	    
	}
	/* BACKGROUND COLOR ====================================== */
	else if ( sscanf( buffer, "b %lf", &bg ) == 1 )
	{
	    fprintf(stderr,"background ignored \n");
	}
	/* COMMENT =============================================== */
	else if ( buffer[0] == '#' )
	{
	}
	/* ERROR================================================== */
	else
	{
	    fprintf(stderr, "Error analyzing line %d\n", line );
	}
    } 
    
    if (fclose(fin) != NULL);
     printf(stderr,"failed to close the input file \n");

    /* ---------- fout = fout + nprims + finter -----------------------*/
    fprintf(fout,"%d\n",nPrims);
    
    fseek(finter,0,0);    /* finter pointe sur le debut du fichier intermediaire */
    while (fgets(buffer,BUFSIZE,finter))
       fprintf(fout,buffer);
    
    fprintf(stderr,"Nombre d'echantillons : %d \n",nSamples);
       
    if (fclose(finter) != NULL)
       printf(stderr,"failed to close the output file \n");
    sprintf(commande,"/bin/rm %s", nfinter);
    system(commande);
    if (fclose(fout) != NULL)
       printf(stderr,"failed to close the output file \n");
          
}



