
/******************************************************************************/
/*			      	       					      */
/*		Module contenant les procedures necessaires		      */
/*		a la mise en oeuvre de Cook-Torrance			      */
/*		Version Analyse Spectrale				      */
/*									      */
/*        	P.Tellier   05/88           E.Langueno 91                     */
/*                                                              	      */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include <math.h>

extern char * memcpy();
extern        qsort();
extern        fclose();
extern double pow();
/******************************************************************************/
#define         VRAI            1  
#define         FAUX            0  
#define         WLMAX           780
#define         WLMIN           380
#define         LGID            20
#define         SPECSIZE        4


#define CLR_SAMPLE_MEYER        0
#define CLR_SAMPLE_HALL         1

/* common geometric constructs
 */
typedef struct {double  r, g, b;}       CLR_RGB;
typedef struct {double  x, y, z;}       CLR_XYZ;

#define RED         0
#define GREEN       1
#define BLUE        2
#define WHITE       3
#define LOAD_MAT(a,b) \
{   int ii,jj; \
    for (ii=0; ii<=2; ii++) \
	for (jj=0; jj<=2; jj++) a[ii][jj] = b[ii][jj]; \
}

       double       D_1[3][3];
static int          min_wl = 380;
static int          max_wl = 780;
static double       xyz_rgb_mat[3][3];
static double       rgb_xyz_mat[3][3];
static double       *X_tristim = NULL;
static double       *Y_tristim = NULL;
static double       *Z_tristim = NULL;
static double       *work_curve = NULL;
static double       xyz_scale = 1.0;
static CLR_XYZ      rgb_primary[4];
static int          clr__cspace_to_xyz ();
static CLR_XYZ      white;


/* This is the RGB to YIQ matrix and YIQ to RGB matrix as
 *  given in Sect 5.2 NTSC and RGB Video
 */
static double       rgb_yiq_mat[3][3] = {
    {0.299,     0.587,      0.144},
    {0.596,    -0.275,     -0.321},
    {0.212,    -0.528,      0.311}};

static double       yiq_rgb_mat[3][3] = {
    {1.0000,    0.9557,     0.6199},
    {1.0000,   -0.2716,    -0.6469},
    {1.0000,   -1.1082,     1.7051}};

/* This is the NTSC primaries with D6500 white point for use as
 *  the default initialization as given in sect 5.1.1 Color
 *  Correction for Display.
 */
static CLR_XYZ      rgb_NTSC[4] = {
    {0.670,     0.330,      0.000},     /* red */
    {0.210,     0.710,      0.080},     /* green */
    {0.140,     0.080,      0.780},     /* blue */
    {0.313,     0.329,      0.358}};    /* white */

/* This is the data for the CIEXYZ curves take from Judd and
 *  Wyszecki (1975), table 2.6, these are for the 1931 standard
 *  observer with a 2-degree visual field.
 */

static double       CIEXYZ[81][4] = {
    {380, 0.0014, 0.0000, 0.0065}, {385, 0.0022, 0.0001, 0.0105},
    {390, 0.0042, 0.0001, 0.0201}, {395, 0.0076, 0.0002, 0.0362},
    {400, 0.0143, 0.0004, 0.0679}, {405, 0.0232, 0.0006, 0.1102},
    {410, 0.0435, 0.0012, 0.2074}, {415, 0.0776, 0.0022, 0.3713},
    {420, 0.1344, 0.0040, 0.6456}, {425, 0.2148, 0.0073, 1.0391},
    {430, 0.2839, 0.0116, 1.3856}, {435, 0.3285, 0.0168, 1.6230},
    {440, 0.3483, 0.0230, 1.7471}, {445, 0.3481, 0.0298, 1.7826},
    {450, 0.3362, 0.0380, 1.7721}, {455, 0.3187, 0.0480, 1.7441},
    {460, 0.2908, 0.0600, 1.6692}, {465, 0.2511, 0.0739, 1.5281},
    {470, 0.1954, 0.0910, 1.2876}, {475, 0.1421, 0.1126, 1.0419},
    {480, 0.0956, 0.1390, 0.8310}, {485, 0.0580, 0.1693, 0.6162},
    {490, 0.0320, 0.2080, 0.4652}, {495, 0.0147, 0.2586, 0.3533},
    {500, 0.0049, 0.3230, 0.2720}, {505, 0.0024, 0.4073, 0.2123},
    {510, 0.0093, 0.5030, 0.1582}, {515, 0.0291, 0.6082, 0.1117},
    {520, 0.0633, 0.7100, 0.0782}, {525, 0.1096, 0.7932, 0.0573},
    {530, 0.1655, 0.8620, 0.0422}, {535, 0.2257, 0.9149, 0.0298},
    {540, 0.2904, 0.9540, 0.0203}, {545, 0.3597, 0.9803, 0.0134},
    {550, 0.4334, 0.9950, 0.0087}, {555, 0.5121, 1.0000, 0.0057},
    {560, 0.5945, 0.9950, 0.0039}, {565, 0.6784, 0.9786, 0.0027},
    {570, 0.7621, 0.9520, 0.0021}, {575, 0.8425, 0.9154, 0.0018},
    {580, 0.9163, 0.8700, 0.0017}, {585, 0.9786, 0.8163, 0.0014},
    {590, 1.0263, 0.7570, 0.0011}, {595, 1.0567, 0.6949, 0.0010},
    {600, 1.0622, 0.6310, 0.0008}, {605, 1.0456, 0.5668, 0.0006},
    {610, 1.0026, 0.5030, 0.0003}, {615, 0.9384, 0.4412, 0.0002},
    {620, 0.8544, 0.3810, 0.0002}, {625, 0.7514, 0.3210, 0.0001},
    {630, 0.6424, 0.2650, 0.0000}, {635, 0.5419, 0.2170, 0.0000},
    {640, 0.4479, 0.1750, 0.0000}, {645, 0.3608, 0.1382, 0.0000},
    {650, 0.2835, 0.1070, 0.0000}, {655, 0.2187, 0.0816, 0.0000},
    {660, 0.1649, 0.0610, 0.0000}, {665, 0.1212, 0.0446, 0.0000},
    {670, 0.0874, 0.0320, 0.0000}, {675, 0.0636, 0.0232, 0.0000},
    {680, 0.0468, 0.0170, 0.0000}, {685, 0.0329, 0.0119, 0.0000},
    {690, 0.0227, 0.0082, 0.0000}, {695, 0.0158, 0.0057, 0.0000},
    {700, 0.0114, 0.0041, 0.0000}, {705, 0.0081, 0.0029, 0.0000},
    {710, 0.0058, 0.0021, 0.0000}, {715, 0.0041, 0.0015, 0.0000},
    {720, 0.0029, 0.0010, 0.0000}, {725, 0.0020, 0.0007, 0.0000},
    {730, 0.0014, 0.0005, 0.0000}, {735, 0.0010, 0.0004, 0.0000},
    {740, 0.0007, 0.0002, 0.0000}, {745, 0.0005, 0.0002, 0.0000},
    {750, 0.0003, 0.0001, 0.0000}, {755, 0.0002, 0.0001, 0.0000},
    {760, 0.0002, 0.0001, 0.0000}, {765, 0.0001, 0.0000, 0.0000},
    {770, 0.0001, 0.0000, 0.0000}, {775, 0.0001, 0.0000, 0.0000},
    {780, 0.0000, 0.0000, 0.0000} };



typedef struct { double wl,val; }  elntcourbe;

typedef elntcourbe Tcourbe[WLMAX-WLMIN+1];

typedef double spectral[SPECSIZE];

/* longueurs d'onde selectionnees (nm) ----------------------*/

double wl[SPECSIZE+1]={456.4, 490.9, 557.7, 631.4, 589.0};

/************************************************* ************/

/* matrice de passage de XYZ a AC1C2 -------------------------*/

double	XYZAC1C2[3][3] = { -0.0177 , 1.0090 , 0.0073 ,
			       -1.5370 , 1.0821 , 0.3209 ,
			       0.1946 ,-0.2045 , 0.5264 };

/* matrice de passage de AC1C2 a XYZ------------------------- */

double	AC1C2XYZ[3][3];
double	XYZ2XYZ[3][3]={1.0/3.84017, 0.0, 0.0,
			   0.0, 1.0/3.84017, 0.0,
			   0.0, 0.0, 1.0/3.84017 };
double   AC1C2XYZ2[3][3];
/* matrice de passage de XYZ a RVB--------------------------- */
/* et sa matrice inverse	   */
/* pour le BARCO CD 351            */

double	XYZRVB[3][3],
    RVBXYZ[3][3];

/* matrice de passage de AC1C2 a RVB-------------------------- */

double AC1C2RVB[3][3];


/******************************************************/
/* 	Correction Gamma			      */
/******************************************************/

#define	GAMMA	1.0/2.3

void correction_gamma(r,v,b)
     double	*r,*v,*b;
{
    *r = pow(*r , (double)GAMMA);
    *v = pow(*v , (double)GAMMA);
    *b = pow(*b , (double)GAMMA);
}

/******************************************************/
/* passage du spectre en coordonnees AC1C2 */

void conversion_ac1c2(spectre,A,C1,C2)
     spectral spectre;
     double *A,*C1,*C2;
{
    
    *A  = 0.18892 * spectre[1] + 0.67493 * spectre[2] + 0.19253 * spectre[3];
    *C1 = 0.31824 * spectre[1] - 0.46008 * spectre[3];
    *C2 = 0.54640 * spectre[0];
}
/******************************************************/
/* passage des coordonnees AC1C2 aux coordonnees RVB  */
/******************************************************/
void conversion_rvb(A,C1,C2,R,V,B)
     double A,C1,C2;
     double *R,*V,*B;
{	
    *R = A * AC1C2RVB[0][0] + C1 * AC1C2RVB[0][1] + C2 * AC1C2RVB[0][2];
    *V = A * AC1C2RVB[1][0] + C1 * AC1C2RVB[1][1] + C2 * AC1C2RVB[1][2];
    *B = A * AC1C2RVB[2][0] + C1 * AC1C2RVB[2][1] + C2 * AC1C2RVB[2][2];
}

/******************************************************/
/* inversion de matrice 3x3                            */
/******************************************************/
void invertmat(m,n)
     double m[3][3];
     double n[3][3];
{
    double           detm, divdetm;
    double           cf[3][3];
    short            i, j;
    
    detm = m[0][0] * (m[1][1] * m[2][2] - m[1][2] * m[2][1]) -
	m[0][1] * (m[1][0] * m[2][2] - m[1][2] * m[2][0]) +
	    m[0][2] * (m[1][0] * m[2][1] - m[1][1] * m[2][0]);
    if (detm != 0.0)
	divdetm = 1.0 / detm;
    else
	(void)printf("matrice non inversible\n");
    
    cf[0][0] = m[1][1] * m[2][2] - m[1][2] * m[2][1];
    cf[0][1] = -(m[1][0] * m[2][2] - m[1][2] * m[2][0]);
    cf[0][2] = m[1][0] * m[2][1] - m[1][1] * m[2][0];
    
    cf[1][0] = -(m[0][1] * m[2][2] - m[0][2] * m[2][1]);
    cf[1][1] = m[0][0] * m[2][2] - m[0][2] * m[2][0];
    cf[1][2] = -(m[0][0] * m[2][1] - m[0][1] * m[2][0]);
    
    cf[2][0] = m[0][1] * m[1][2] - m[0][2] * m[1][1];
    cf[2][1] = -(m[0][0] * m[1][2] - m[0][2] * m[1][0]);
    cf[2][2] = m[0][0] * m[1][1] - m[0][1] * m[1][0];
    
    for (i = 0; i < 3; i++)
	for (j = 0; j < 3; j++)
	    n[i][j] = divdetm * cf[j][i];
}
/******************************************************/
/* produit de 2 matrices 3x3 */

void multmatrice(a,b,c)
     double a[3][3],b[3][3],c[3][3];
{
    short i,j,k;
    
    for (i=0;i<3;i++){
	for (j=0;j<3;j++){
	    for (k=0;k<3;k++){
		c[i][j] += a[i][k] * b[k][j];
	    }
	}
    }
}

/******************************************************/
/* initialisation de la matrice de passage de RVB a XYZ */
/* cf GREENBERG						*/

void init_matrix(mat)
     double mat[3][3];
{
    double	rx,ry,rz,gx,gy,gz,bx,by,bz;
    double  wx,wy;
    double 	Cr,Cg,Cb;
    double	D;
    
    /* pour la console BARCO */
    
    wx = 0.3127;   /* -- */
    wy = 0.3290;   /* -- */
    
    rx = 0.64;     /* -- */
    ry = 0.33;     /* -- */
    gx = 0.29;     /* -- */
    gy = 0.60;     /* -- */
    bx = 0.15;     /* -- */
    by = 0.06;     /* -- */
    
    /* pour la console BARCO */
    
    rz = 1.0 - rx - ry;
    gz = 1.0 - gx - gy;
    bz = 1.0 - bx - by;
    
    Cr = wx * (gy - by) - wy * (gx - bx) + gx * by - bx * gy;
    Cg = wx * (by - ry) - wy * (bx - rx) - rx * by + bx * ry;
    Cb = wx * (ry - gy) - wy * (rx - gx) + rx * gy - gx * ry;
    D = rx * (gy - by) + gx * (by - ry) + bx * (ry - gy);
    
    Cr /= D;
    Cg /= D;
    Cb /= D;
    
    mat[0][0] = rx * Cr;
    mat[0][1] = gx * Cg;
    mat[0][2] = bx * Cb;
    mat[1][0] = ry * Cr;
    mat[1][1] = gy * Cg;
    mat[1][2] = by * Cb;
    mat[2][0] = rz * Cr;
    mat[2][1] = gz * Cg;
    mat[2][2] = bz * Cb;
    
}
/******************************************************/			
/* delivre le flux d'un patch */
/******************************************************/

double flux(spec,S)
     spectral spec;
     double   S;
{
    double leflux;
    leflux = spec[0]*97.0 + spec[1]*50.0 + spec[2]*74.0 + spec[3]*93.0;
    /* should be multiplied by S */
    return(leflux);
    
}

/******************************************************/
/* compare 2 elnt d'un spectre sur leur longueur d'onde */

int compwl(a, b)
     elntcourbe *a, *b;
{
    return( (int)(a->wl - b->wl));
    /* OK si echantillons distants d'au moins 1 nm */
}

/******************************************************/
/* preleve les echantillons de meyer du spectre tab 
   et les met dans spect */
/******************************************************/
void echantMeyer(tab, nb, wl, spect, f0)
     Tcourbe tab;
     int nb;
     double wl[SPECSIZE+1];
     spectral spect;
     double  *f0;
{
    int     i,j;
    double     lambda;
    double spec[SPECSIZE+1];
    
    for (i=0;i<SPECSIZE+1;i++)
	{
            lambda = wl[i];
            if (lambda <= tab[0].wl)
                spec[i]=tab[0].val-(tab[1].val-tab[0].val)*(tab[0].wl-lambda)
		    / (tab[1].wl - tab[0].wl);
            else {
                if (lambda >= tab[nb].wl)
                    spec[i]=tab[nb-1].val+(tab[nb].val-tab[nb-1].val)
			*(lambda-tab[nb].wl)/(tab[nb].wl-tab[nb-1].wl);
                else {  /* recherche des valeurs qui encadrent lambda */
                    for (j=0;j<=nb;j++)
                        if (tab[j].wl > lambda)
                            break;
                    spec[i]=tab[j-1].val+(tab[j].val-tab[j-1].val)
                        *(lambda-tab[j-1].wl)/(tab[j].wl - tab[j-1].wl);
		}
	    }
	}
    (void)memcpy((char *)spect, (char *)spec, sizeof(spectral));
    *f0 = spec[SPECSIZE];
}

/* pompe a Roy HALL !!! */

/* ****************************************************************
 * CLR_read_mtl (file, opaque, n, k, curve, f0lu)
 *  char        *file       (in) material file name
 *  int         *conduct    (out) conductor - set to FAUX if not
 *                              specified in the material file
 *  double      *n, *k      (out) n and k, - set to 0.0 if not
 *                              specified in the material file
 *  spectral     curve      (out) curve array
 *
 * Returns VRAI if the material was read, FAUX if the material
 *  could not be found or an error was detected in the material
 *  file.  NULL pointers can be given as arguments for 'conduct',
 *  'n', 'k', and/or 'curve' if the properties are not of interest.
 *
 * The material file must be in the format given in Sect III.7.1
 *  Material Data.
 */
/******************************************************/
int CLR_read_mtl (file, opaque, n, k, curve, f0lu)
     char        *file;
     int         *opaque;
     double         *n, *k;
     spectral    curve;
     double      *f0lu;
{
    FILE        *fp1, *fopen();
    char        in_str[200];
    int         nbval=0;
    Tcourbe     courbe;
    char        *dir;
    char        name[80];

    dir = "/bd2/insa5/etudiants/DESS_IN/projdess/bdphoto/";
    strcpy(name,"");
    strcat(name,dir);
    strcat(name,file);

    if ((fp1=fopen(name,"r")) == NULL) 
	{
	    printf("can't open %s\n",name);exit(0);
	}
    if (n != NULL) *n = 0.0;
    if (k != NULL) *k = 0.0;
    if (opaque != NULL) *opaque = VRAI;
    while (fgets(in_str,200,fp1))
	{
	    switch (in_str[0])
		{
		  case '#':       
		    break;
		  case 'k':       
		    if (k != NULL) (void)sscanf (in_str, "k %lf", k);
		    break;
		  case 'n':      
		    if (n != NULL) (void)sscanf (in_str, "n %lf", n);
		    break;
		  case 'O':    
		    if (opaque != NULL) *opaque = VRAI;
		    break;
		  case 'T':      
		    if (opaque != NULL) *opaque = FAUX;
		    break;
		  default:       
		    if ((sscanf(in_str,
				"%lf %lf", &courbe[nbval].wl, &courbe[nbval].val) == 2))
			nbval++;
		} 
	}
    if((k!=NULL)&&(*k!=0.0)) *opaque=VRAI;
    
    (void)qsort((char *)courbe, nbval, sizeof(elntcourbe), compwl);
    
    (void)echantMeyer(courbe, nbval-1, wl, curve, f0lu);  
    /*fprintf(stderr,"%s %g %g %g %g\n",file,curve[0],curve[1],curve[2],curve[3]);*/
    fclose(fp1);
    return VRAI;
}
/*----------------------base de fonctions----------------------------------*/
double F1(la)
     double la;
{
    return ( (double)1.0);
}

double F2(la)
     double la;
{
    return ( (double)sin(2*PI*((la -380)/400.0)));
}

double F3(la)
     double la;
{
    return ( (double)cos(2*PI*((la -380)/400.0)));
}
/*----------------------------------------------------------------*/
/* ****************************************************************
 * double CLR_area_spect (c1)
 *  double  *c1         (in)  - input curve
 *
 * Returns the are under the spectral curve 'c1'.  Returns 0 if
 *  the CLR_ routines are not initialized.
 */
double CLR_area_spect (c1)
double  *c1;
{   int     ct;
    double  area = 0.0;

    for (ct=max_wl-min_wl+1; --ct>=0; ) area += *c1++;
    return area;
}
/*----------------------------------------------------------------*/
/* ****************************************************************
 * clr__cspace_to_xyz (cspace, t_mat)
 *  CLR_XYZ       cspace[4]   (in)  - the color space definition,
 *                                      3 primaries and white
 *  double        t_mat[3][3] (mod) - the color transformation
 *
 * Builds the transformation from a set of primaries to the CIEXYZ
 *  color space.  This is the basis for the generation of the color
 *  transformations in the CLR_ routine set.  The method used is
 *  that detailed in Sect 3.2 Colorimetry and the RGB monitor.
 *  Returns FAUX if there is a singularity.
 */
static int clr__cspace_to_xyz (cspace, t_mat)
CLR_XYZ     cspace[];
double      t_mat[][3];
{   int     ii, jj, kk, tmp_i, ind[3];
    double  mult, white[3], scale[3];

    /* normalize the white point to Y=1 */
    if (cspace[WHITE].y <= 0.0) return FAUX;
    white[0] = cspace[WHITE].x / cspace[WHITE].y;
    white[1] = 1.0;
    white[2] = cspace[WHITE].z / cspace[WHITE].y;

    for (ii=0; ii<=2; ii++) {
	t_mat[0][ii] = cspace[ii].x;
	t_mat[1][ii] = cspace[ii].y;
	t_mat[2][ii] = cspace[ii].z;
	ind[ii] = ii;
    }

    /* gaussian elimination  with partial pivoting */
    for (ii=0; ii<2; ii++) {
	for (jj=ii+1; jj<=2; jj++)
	    if (fabs(t_mat[ind[jj]][ii]) >
		    fabs(t_mat[ind[ii]][ii])) {
		tmp_i=ind[jj]; ind[jj]=ind[ii]; ind[ii]=tmp_i;
	    }
	if (t_mat[ind[ii]][ii] == 0.0) return FAUX;

	for (jj=ii+1; jj<=2; jj++) {
	    mult = t_mat[ind[jj]][ii] / t_mat[ind[ii]][ii];
	    for (kk=ii+1; kk<=2; kk++)
		t_mat[ind[jj]][kk] -= t_mat[ind[ii]][kk] * mult;
	    white[ind[jj]] -= white[ind[ii]] * mult;
	}
    }
    if (t_mat[ind[2]][2] == 0.0) return FAUX;

    /* back substitution to solve for scale */
    scale[ind[2]] = white[ind[2]] / t_mat[ind[2]][2];
    scale[ind[1]] = (white[ind[1]] - (t_mat[ind[1]][2] *
	scale[ind[2]])) / t_mat[ind[1]][1];
    scale[ind[0]] = (white[ind[0]] - (t_mat[ind[0]][1] *
	scale[ind[1]]) - (t_mat[ind[0]][2] * scale[ind[2]])) /
	t_mat[ind[0]][0];

    /* build matrix */
    for (ii=0; ii<=2; ii++) {
	t_mat[0][ii] = cspace[ii].x * scale[ii];
	t_mat[1][ii] = cspace[ii].y * scale[ii];
	t_mat[2][ii] = cspace[ii].z * scale[ii];
    }

    return VRAI;
}

/*-----------------------------------*/
int CLR_init (min, max, rgb)
int         min;
int         max;
CLR_XYZ     rgb[4];
{   int     color, ct;
    double  len;
    char    *malloc();

    min_wl = min;
    max_wl = max;

    /* load primaries and build transformations, use the
     * defaults is rgb==NULL
     */
    for (color=RED; color<=WHITE; color++)
       {if (rgb == NULL) rgb_primary[color] = rgb_NTSC[color];
	else rgb_primary[color] = rgb[color];
	rgb_primary[color].z = 1.0 -
	    rgb_primary[color].x - rgb_primary[color].y;
       }
    if (!clr__cspace_to_xyz(rgb_primary, rgb_xyz_mat)) goto error;
    invertmat(rgb_xyz_mat, xyz_rgb_mat);

    /* build the XYZ sampling curves - allocate space for the
     *  curves and interpolate the CIEXYZ table into continuous
     *  curves at 1nm increments.
     */
   {   int     ii, nm;
	double  *x, *y, *z;
	double  x_cur, y_cur, z_cur;
	double  x_inc, y_inc, z_inc;
	if ((work_curve = (double *)malloc((unsigned)
	    (sizeof(double) * (max_wl-min_wl+1))))
	    == NULL) goto error;
	if ((x = X_tristim = (double *)malloc((unsigned)
	    (sizeof(double) * (max_wl-min_wl+1))))
	    == NULL) goto error;
	if ((y = Y_tristim = (double *)malloc((unsigned)
	    (sizeof(double) * (max_wl-min_wl+1))))
	    == NULL) goto error;
	if ((z = Z_tristim = (double *)malloc((unsigned)
	    (sizeof(double) * (max_wl-min_wl+1))))
	    == NULL) goto error;
	for (nm=min_wl; nm<380; nm++) *x++ = *y++ = *z++ = 0.0;
	for (ct=0; ct<80; ct++)
           {x_cur = CIEXYZ[ct][1];
	    y_cur = CIEXYZ[ct][2];
	    z_cur = CIEXYZ[ct][3];
	    x_inc = (CIEXYZ[ct+1][1] - x_cur) / 5.0;
	    y_inc = (CIEXYZ[ct+1][2] - y_cur) / 5.0;
	    z_inc = (CIEXYZ[ct+1][3] - z_cur) / 5.0;
	    for (ii=0; ii<5; ii++, nm++) 
               {if (nm > max_wl) goto XYZ_done;
		if (nm >= min_wl)
                   {*x++ = x_cur; *y++ = y_cur; *z++ = z_cur;
		   }
		x_cur += x_inc; y_cur += y_inc; z_cur += z_inc;
	        }
	    }
	if (nm <= max_wl)
           {*x++ = x_cur; *y++ = y_cur; *z++ = z_cur; nm++;
	   }
	for ( ;nm<=max_wl; nm++) *x++ = *y++ = *z++ = 0.0;
    }
XYZ_done:
    /* determine the scaling factor to be used in sampling spectral
     *  curves to bring Y os a sampled identity curve to 1.
     */
    xyz_scale = 1.0 / CLR_area_spect (Y_tristim);

    return VRAI;

error:
    CLR_exit();
    return FAUX;
}

/* ****************************************************************
 * CLR_exit()
 *
 * Completes use of the CLR_ routine set and frees any
 *   allocated space
 */
CLR_exit()
{
    /*if (!init) return VRAI;
    (void)CLR_exit_samples();*/
    if (X_tristim != NULL) free((char *)X_tristim);
    if (Y_tristim != NULL) free((char *)Y_tristim);
    if (Z_tristim != NULL) free((char *)Z_tristim);
    if (work_curve != NULL) free((char *)work_curve);
    X_tristim = Y_tristim = Z_tristim = work_curve = NULL;
    return VRAI;
}

/*----------------------------------------------------------------*/
void genere_D_1(D)
     double D[3][3];    
{
 double temp1,temp2,temp3;
 int i;
 double D2[3][3];
 double area;
 
 /* attention clr _init a faire avant  */
 
area=CLR_area_spect (Y_tristim);
 
 temp1=temp2=temp3=(double)0.0;
 for(i=380;i<=780;i++)
     {
	 temp1+=F1((double)i)*X_tristim[i-380];
         temp2+=F1((double)i)*Y_tristim[i-380];
         temp3+=F1((double)i)*Z_tristim[i-380];
      }
 temp1/=area;
 temp2/=area;
 temp3/=area;
 
  D2[0][0] = temp1 * XYZRVB[0][0] + temp2 * XYZRVB[0][1] + temp3 * XYZRVB[0][2];
  D2[1][0] = temp1 * XYZRVB[1][0] + temp2 * XYZRVB[1][1] + temp3 * XYZRVB[1][2];
  D2[2][0] = temp1 * XYZRVB[2][0] + temp2 * XYZRVB[2][1] + temp3 * XYZRVB[2][2];

 temp1=temp2=temp3=(double)0.0;
 for(i=380;i<=780;i++)
     {
	 temp1+=F2((double)i)*X_tristim[i-380];
         temp2+=F2((double)i)*Y_tristim[i-380];
         temp3+=F2((double)i)*Z_tristim[i-380];
      }
 temp1/=area;
 temp2/=area;
 temp3/=area;
 
  D2[0][1] = temp1 * XYZRVB[0][0] + temp2 * XYZRVB[0][1] + temp3 * XYZRVB[0][2];
  D2[1][1] = temp1 * XYZRVB[1][0] + temp2 * XYZRVB[1][1] + temp3 * XYZRVB[1][2];
  D2[2][1] = temp1 * XYZRVB[2][0] + temp2 * XYZRVB[2][1] + temp3 * XYZRVB[2][2];

 temp1=temp2=temp3=(double)0.0;
 for(i=380;i<=780;i++)
     {
	 temp1+=F3((double)i)*X_tristim[i-380];
         temp2+=F3((double)i)*Y_tristim[i-380];
         temp3+=F3((double)i)*Z_tristim[i-380];
      }
 temp1/=area;
 temp2/=area;
 temp3/=area;
 
  D2[0][2] = temp1 * XYZRVB[0][0] + temp2 * XYZRVB[0][1] + temp3 * XYZRVB[0][2];
  D2[1][2] = temp1 * XYZRVB[1][0] + temp2 * XYZRVB[1][1] + temp3 * XYZRVB[1][2];
  D2[2][2] = temp1 * XYZRVB[2][0] + temp2 * XYZRVB[2][1] + temp3 * XYZRVB[2][2];

 invertmat(D2,D);
 
}

/*----------------------------------------------------------------*/
void genere_spectre4L(R,G,B,spectre)
     double R,G,B;
     spectral spectre;
 
{  int i;
   double W1,W2,W3;
 
  W1 = R * D_1[0][0] + G * D_1[0][1] + B * D_1[0][2];
  W2 = R * D_1[1][0] + G * D_1[1][1] + B * D_1[1][2];
  W3 = R * D_1[2][0] + G * D_1[2][1] + B * D_1[2][2];
  for(i=0;i<=3;i++)
    {  
     spectre[i]=W1*F1(wl[i])+W2*F2(wl[i])+W3*F3(wl[i]);
    }
   
}


/*----------------------------------------------------------------*/
void genere_spectre(R,G,B,curve)
     double R,G,B;
     Tcourbe curve;
 
{  int i;
   double W1,W2,W3;
   FILE *fd;
 
  W1 = R * D_1[0][0] + G * D_1[0][1] + B * D_1[0][2];
  W2 = R * D_1[1][0] + G * D_1[1][1] + B * D_1[1][2];
  W3 = R * D_1[2][0] + G * D_1[2][1] + B * D_1[2][2];
  /* printf("w1: %g   w2: %g   w3: %g\n",W1,W2,W3);*/
   fd=fopen("ficout","w");   
  for(i=380;i<=780;i++)
    {  
	curve[i-380].wl=(double)i;
	curve[i-380].val=W1*F1((double)i)+W2*F2((double)i)+W3*F3((double)i);
        fprintf(fd,"%g %g\n",curve[i-380].wl,curve[i-380].val);
    }
   fclose(fd);
   
}


/*----------------------------------------------------------------*/
/* pour transformer des  triplets rgb en spectres(380..780) il faut:

    -  CLR_init (380, 780, NULL);   une fois pour toutes
    -  genere_D_1(D_1);             une fois pour toutes
    -  genere_spectre(R,G,B,curve); pour chaque triplet
                           avec "Tcourbe curve"
    -  (void)echantMeyer(curve, 400, wl,spectre, f0lu);
                      echantillonne sur quatres longueurs d'onde


  pour transformer directement le triplet rgb en spectre 4 echantillons
    -  genere_spectre4L(R,G,B,spectre);  avec "double  spectre[3]"
*/
/*----------------------------------------------------------------*/
NormMatrix(M)
     double M[3][3];
{
    int i, j;
    double sigma;

    sigma = M[1][0] + M[1][1] +  M[1][2];

    for(i=0;i<3;i++)
    {
        for(j=0;j<3;j++)
        {
            M[i][j] /= sigma;
        }
    }
}


/******************************************************/
/* Initialisation des diverses matrices de transformation */
/* ( a faire en debut de prog) */
/******************************************************/

void init_matrices()
{
    invertmat(XYZAC1C2,AC1C2XYZ); 
    init_matrix(RVBXYZ);
    NormMatrix(RVBXYZ);
    invertmat(RVBXYZ,XYZRVB);
    multmatrice(XYZRVB,AC1C2XYZ,AC1C2RVB);
    CLR_init (380, 780, NULL);
    genere_D_1(D_1);
}

/*----------------------------------------------------------------*/


/* a faire pour chaque pixel  . A C1 C2 ... sont des doubles  */
/*
===conversion_ac1c2(color,&A,&C1,&C2);
===conversion_rvb(A,C1,C2,&R,&V,&B);
===correction_gamma(&R,&V,&B);


===main() 
==={  spectral curve;
===   double n,k,l,f0lu;
===
===  init_matrices();
===  CLR_read_mtl ("brique_jaune", &l, &k, &n, curve, &f0lu);
===   printf("%f %f %f %f\n",curve[0],curve[1],curve[2],curve[3]);
===}
===
*/



